STATUS = {
    200:"OK",

    400:"Bad request",
    451:"Message must be a dictionary",
    452:"Missing 'action' Field",
    453:"Name not allowed",
    454:"'name' Already Subscribed",
    455:"Missing 'to' Field",
    456:"Client Not Subscribed",
    457:"Missing 'msg' Field",
    458:"'subscribe' is the only First Action Allowed",
    459:"Unknown 'action' value"
    }


from . import utils
from . import interface as sc
import socket, json


class Client():
    def __init__(self, IP, PORT):
        self.IP = IP
        self.PORT = PORT
        self.socket = None

    def start(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(1)
        self.socket.connect((self.IP, self.PORT))
        self.socket.settimeout(None)

    def send(self, data, key=None):
        data = json.dumps(data)
        sc.msg_send(self.socket, data)
    
    def recv(self, key=None):
        try:
            data = sc.msg_recv(self.socket)
            data = json.loads(data)
        except json.decoder.JSONDecodeError:
            utils.debug(data,"json Decode Error")
            return ""
        return data

    def close(self):
        if socket:
            self.socket.close()


import struct
from sys import getsizeof


"""
PACKET STRUCTURE

1Byte + NByte + CByte
  N   +   C   + data

Max N value = (2^8) = 256
Max N value = (2^(256*8)) = (2^(2048))
Max data byte = (2^(2048))

"""


BYTES_OBJECT_OVERHEAD = getsizeof(b"")

types = {
    "<B":1,
    "<H":2,
    "<L":4,
    "<Q":8
}




class msg_send_Exception(Exception):
    def __init__(self, message="Error"):
        super().__init__(message)

class msg_recv_Exception(Exception):
    def __init__(self, message="Error"):
        super().__init__(message)



def msg_recv(socket):
    try:
        N = socket.recv(1)
    except ConnectionResetError as e:
        raise e
    except ConnectionAbortedError as e:
        raise e
    except OSError as e:
        raise e
    
    if N == b"":
        raise msg_recv_Exception("Socket closed")
    N = struct.unpack('<B', N)[0]

    if N != 0:
        C = recvNBytes(socket, N)
        for t in types:
            if types[t] == N:
                C = struct.unpack(t, C)[0]
        
        if C != 0:
            data = recvNBytes(socket, C)
            data = data.decode('utf-8')
        else:
            raise msg_recv_Exception("Error reciving C")
    
    else:
        raise msg_recv_Exception("Error reciving N")
    
    return data


def msg_send(socket, data):
    if type(data) is not str:
        raise msg_send_Exception("'data' must be string")
    if data == "":
        raise msg_send_Exception("'data' must not be empty")

    data = data.encode("utf-8")
    C = int(len(data))
    N = (C.bit_length()+7)//8

    if N>8:
        raise msg_send_Exception("Message too long")

    for t in types:
        if N<=types[t]:
            socket.send(struct.pack('<B', types[t]))
            socket.send(struct.pack(t, C))
            socket.send(data)
            return None

    raise msg_send_Exception("Error in msg_send")
    

def recvNBytes(socket, N):
    data = socket.recv(N)
    while getsizeof(data)-BYTES_OBJECT_OVERHEAD!=N:
        data += socket.recv(N-getsizeof(data)+BYTES_OBJECT_OVERHEAD)
    return data


def msg_encode(data, key):
    pass


def msg_decode(data, key):
    pass




import socket, json, time
from . import utils
from . import interface as sc
from .SThread import SThread
import uuid

notAllowedNames = ['', 'all']

class Server():
    def __init__(self, IP, PORT):
        self.IP = IP
        self.PORT = PORT
        self.socket = None
        self.connected_clients = {}

    def start(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((self.IP, self.PORT))
        self.socket.listen()
        self.server_loop()
    
    def send(self, connection, data, key=None):
        data = json.dumps(data)
        utils.debug("send:",data)
        sc.msg_send(connection, data)
    
    def recv(self, connection, key=None):
        try:
            data = sc.msg_recv(connection)
            utils.debug("recv:",data)
        except ConnectionResetError as e:
            return ""
        except ConnectionAbortedError as e:
            return ""
        except OSError as e:
            return ""
        
        try:
            data = json.loads(data)
        except json.decoder.JSONDecodeError:
            utils.debug("Error",data,"json Decode Error")
        return data
    
    def server_loop(self):
        while True:
            conn, addr = self.socket.accept()
            client_thread = SThread(target=self.client_startup)
            client_thread._args=(conn, client_thread)
            client_thread.start()

    def client_startup(self, conn, thread):
        data = self.recv(conn)
        if data=="":
            exit(0)
        if not self.actionCheck(conn, data):
            self.closeSocket(conn)
            return

        if data["action"] == "subscribe":
            self.subscribeAction(conn, data, thread)
            utils.debug("clients:",self.connected_clients)
        else:
            self.sendStatus(conn, 458)
            self.closeSocket(conn)
            return
    
    def handle(self, conn, name):
        while True:
            data = self.recv(conn)
            if (data==""):
                self.removeClient(name)
                exit(0)
            if not self.actionCheck(conn, data):
                self.removeClient(name)
                exit(0)

            if data["action"] == "send":
                self.sendAction(conn, data, name)
            
            elif data["action"] == "getClients":
                self.getClientsAction(conn)
            
            elif data["action"] == "close":
                self.closeAction(conn, data)
                exit(0)
            
            else:
                self.unknownAction(conn, name)

    def actionCheck(self, conn, data):
        if type(data) is not dict:
            self.sendStatus(conn, 451)
            return False
        if "action" not in data:
            self.sendStatus(conn, 452)
            return False
        return True

    def closeSocket(self, conn):
        conn.close()
    
    def removeClient(self, name):
        if name in self.connected_clients:
            self.closeSocket(self.connected_clients[name][0])
            del self.connected_clients[name]
            utils.debug("client",name,"disconnected")

    def stopThread(self, name):
        if name in self.connected_clients:
            thread = self.connected_clients[name][1]
            del self.connected_clients[name]
            thread.stop()

    def subscribeAction(self, conn, data, thread):
        client_name = None
        if "name" in data:
            client_name = data["name"]
            if client_name in self.connected_clients:
                self.sendStatus(conn, 454)
                return
            if client_name in notAllowedNames:
                self.sendStatus(conn, 453)
                return
        else:
            client_name = str(uuid.uuid4())
            while client_name in self.connected_clients:
                client_name = str(uuid.uuid4())

        self.connected_clients[client_name] = ([conn, thread])
        utils.debug("client",client_name,"connected")
        self.sendStatus(conn, 200)
        self.send(conn, {'action':'subscribeName', 'subscribeName':client_name})
        self.handle(conn, client_name)
    
    def closeAction(self, name):
        self.removeClient(name)

    def getClientsAction(self, conn):
        clients_name = [n for n in self.connected_clients]
        self.send(conn, {'action':'clientsList', 'clients':clients_name})

    def sendAction(self, conn, data, name):
        if "to" in data:
            dst_client = data["to"]
            if dst_client in self.connected_clients:
                if "msg" in data:
                    self.send(self.connected_clients[dst_client][0], {'action':'msg', 'msg':data['msg'], 'from':name})
                    self.sendStatus(conn, 200)
                else:
                    self.sendStatus(conn, 457)
            elif dst_client=="all":
                if "msg" in data:
                    for client in self.connected_clients:
                        self.send(self.connected_clients[client][0], {'action':'msg', 'msg':data['msg'], 'from':name})
                    self.sendStatus(conn, 200)
            else:
                self.sendStatus(conn, 456)
        else:
            self.sendStatus(conn, 455)

    def unknownAction(self, conn, data):
        self.sendStatus(conn, 459)

    def sendStatus(self, conn, status):
        self.send(conn, {'action':'status', 'status':status})




import threading

class SThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self,  *args, **kwargs):
        super(SThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()